/******************************************************************************
 *  main.c
 *  pss2wave main source
 *
 *  Created by Decomo@Freyja on  07/10/17.
 *
 * 【更新履歴】
 * 2007-10-17 Version 1.00
 *              ・完成
 *
 *****************************************************************************/
#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>

#include "PKUtilities.h"

/* ////////////////////////////////////////////////////////////////////////////
    定数、構造体定義
*/ ////////////////////////////////////////////////////////////////////////////
//#define DEBUG_PRINT(x) fprintf(stderr, x)
#define DEBUG_PRINT(x)

#define DEFAULT_OUTPUT_FILENAME	"a.wav"

#define PSS_HEADER_TYPES            3
#define PSS_HEADER_SIZE_TYPE1_1		0x3F
#define PSS_HEADER_SIZE_TYPE1_2     0x17

#define WAVE_HEADER_SIZE		44

#define PRIVATE_STREAM_1		-4
#define PACKET_START_CODE		0x0A
#define MPEG_PROGRAM_END_CODE	-1
#define PROGRAM_FAILED          -2

typedef struct {
	FILE *fpInput;
	FILE *fpOutput;
	
	uint32_t interleaveSize;
	uint16_t interleaveCount;
	
	uint32_t samplingRate;
	uint16_t numCh;
	uint32_t totalAudioDataSize;
	uint32_t totalAudioDataCount;
    
    uint8_t *channel;
	uint8_t *chL;
	uint8_t *chR;
} PSS2WAVE;

/* ////////////////////////////////////////////////////////////////////////////
    プロトタイプ宣言
*/ ////////////////////////////////////////////////////////////////////////////
void Initialize(PSS2WAVE *data, const char *inputFile, const char *outputFile);
void Convert(PSS2WAVE *data);
void Release(PSS2WAVE *data);

int32_t checkPSSHeader(PSS2WAVE *data);
int32_t ripAudioData(PSS2WAVE *data);

void writeWAVEHeader(PSS2WAVE *data);
void writeWAVEData(PSS2WAVE *data);

/*! ///////////////////////////////////////////////////////////////////////////
@brief メイン関数
*/ ////////////////////////////////////////////////////////////////////////////
int main (int argc, const char * argv[])
{
	PSS2WAVE data;
	char *inputFile, *outputFile;
    
    // 引数チェック
	switch (argc)
	{
        default:
		case 1:
			printf("usage: pss2wave pss_file [output_file]\n");
			return 0;
		case 2:
			inputFile = argv[1];
            outputFile = DEFAULT_OUTPUT_FILENAME;
			break;
		case 3:
			inputFile = argv[1];
			outputFile = argv[2];
	}
	
    // 初期化
	Initialize(&data, inputFile, outputFile);
	
    // 変換
    Convert(&data);

    // 後始末
    Release(&data);
	
    return 0;
}

/*! ///////////////////////////////////////////////////////////////////////////
@brief PSS2WAVE構造体を初期化し、PSS/WAVEファイルを設定する。
@param[out] data PSS2WAVE構造体へのポインタ。
@param[in]  inputFile 変換するPSSファイル名。
@param[in]  outputFile 出力するWAVEファイル名。
@return なし。
*/ ////////////////////////////////////////////////////////////////////////////
void Initialize(PSS2WAVE *data, const char *inputFile, const char *outputFile)
{
	 if ( (data->fpInput = fopen(inputFile, "rb")) == NULL )
	 {
		 fprintf(stderr, "Can't open the PSS file : %s\n", inputFile);
		 exit(EXIT_FAILURE);
	 }
	
	if ( (data->fpOutput = fopen(outputFile, "wb")) == NULL )
	{
		fprintf(stderr, "Can't open the output file : %s\n", outputFile);
		exit(EXIT_FAILURE);
	}
	
	data->interleaveSize = 0;
	data->interleaveCount = 0;
    data->totalAudioDataCount = 0;
}

/*! ///////////////////////////////////////////////////////////////////////////
@brief 終了処理。PSS2WAVEを開放する。
@param[in] data PSS2WAVE構造体へのポインタ。
@return なし。
*/ ////////////////////////////////////////////////////////////////////////////
void Release(PSS2WAVE *data)
{
	free(data->chL);
	free(data->chR);
	
	fclose(data->fpInput);
	fclose(data->fpOutput);
}

/*! ///////////////////////////////////////////////////////////////////////////
@brief PSSファイルをWAVEファイルに変換する。
@param[in, out] data PSS2WAVE構造体へのポインタ
*/ ////////////////////////////////////////////////////////////////////////////
void Convert(PSS2WAVE *data)
{
    int32_t seek;
    
	while (feof(data->fpInput)==0)
	{
		seek = checkPSSHeader(data);

        if (seek == PROGRAM_FAILED)
        {
            fprintf(stderr, "Convert failed.\n");
            return;
        }
        
		if (seek == MPEG_PROGRAM_END_CODE)
        {
            fprintf(stdout, "complete.\n");
            return;
        }
        
		if (seek == PRIVATE_STREAM_1)
        {
            fseek(data->fpInput, seek, SEEK_CUR);
            ripAudioData(data);
            continue;
        }

        fseek(data->fpInput, seek, SEEK_CUR);
	}
    
    printf("End of file.\n");
}

/*! ///////////////////////////////////////////////////////////////////////////
@brief PSSファイルのヘッダを解析する。
@param[in] data PSS2WAVE構造体へのポインタ。
@return 次にシークすべき値。あるいはプログラム制御コード。
*/ ////////////////////////////////////////////////////////////////////////////
int32_t checkPSSHeader(PSS2WAVE *data)
{
	const uint8_t HeaderType[PSS_HEADER_TYPES][4] = {
		{0x00, 0x00, 0x01, 0xBD},   // Type 1 (private_stream_1: Audio data)
		{0x00, 0x00, 0x01, 0xBA},   // Type 2 (pack_start_code)
		{0x00, 0x00, 0x01, 0xB9}    // Type 3 (MPEG_program_end_code)
	};
	
	unsigned char tmp[5];
	uint32_t i=0;
	uint32_t blockSize;
	
    // ヘッダ識別子の特定
    fread(tmp, 4, 1, data->fpInput);
	while (i<PSS_HEADER_TYPES && memcmp(tmp, &HeaderType[i], 4)!=0)
		i++;
	i++;	// Typeが0から始まるのは気持ち悪いので+1しておく
    
	switch (i)
	{
        case 4:
			fread(tmp, 2, 1, data->fpInput);
			blockSize = BE_ARRAY_TO_INT16(tmp);
			return blockSize;
		case 3:
			return MPEG_PROGRAM_END_CODE;
		case 2:
			return PACKET_START_CODE;
		case 1:
            return PRIVATE_STREAM_1;
    }            

	fprintf(stderr, "** AN EXCEPTION REACHE **\n");
	return PROGRAM_FAILED;
}

/*! ///////////////////////////////////////////////////////////////////////////
@brief PSSファイルから音声を切り出す。
@param[in, out] data PSS2WAVE構造体へのポインタ。
@retval 0 正常終了。
@retval PROGRAM_FAILED 異常終了。
*/ ////////////////////////////////////////////////////////////////////////////
int32_t ripAudioData(PSS2WAVE *data)
{
	unsigned char tmp[5];
	char PSSHeader[PSS_HEADER_SIZE_TYPE1_1];
	uint32_t blockSize, audioSize;
    int i;
    
    // Type 1が初出か否かでヘッダサイズが異なる
    if (data->interleaveSize==0)
        fread(PSSHeader, PSS_HEADER_SIZE_TYPE1_1, 1, data->fpInput);
    else
        fread(PSSHeader, PSS_HEADER_SIZE_TYPE1_2, 1, data->fpInput);
    
    // ブロックサイズ
    memcpy(tmp, &PSSHeader[0x04], 2);
    blockSize = BE_ARRAY_TO_INT16(tmp);
    
    // オーディオデータサイズ（2回目以降用）
    audioSize = blockSize - 0x17 + 0x06;
    
    // オーディオデータの判別
    memcpy(tmp, &PSSHeader[0x14], 1);
    if ( tmp[0]!=0xa0 )
    {
        fprintf(stderr, "Non PCM data.\n");
        return PROGRAM_FAILED;
    }
    
    // Type 1が初めて現れた時
    // + インターリーブサイズなどの設定
    if (data->interleaveSize==0)
    {
        // オーディオデータサイズ（初回のみ）
        audioSize = blockSize - 0x3F + 0x06;
        
        // サンプリングレート
        memcpy(tmp, &PSSHeader[0x23], 4);
        data->samplingRate = LE_ARRAY_TO_INT32(tmp);
        
        // チャンネル数
        memcpy(tmp, &PSSHeader[0x27], 4);
        data->numCh = LE_ARRAY_TO_INT32(tmp);
        
        // インタリーブサイズ
        memcpy(tmp, &PSSHeader[0x2B], 4);
        data->interleaveSize = LE_ARRAY_TO_INT32(tmp);
        
        // 全オーディオデータサイズ
        memcpy(tmp, &PSSHeader[0x3B], 4);
        data->totalAudioDataSize = LE_ARRAY_TO_INT32(tmp);
        
        if ( (data->chL = malloc(sizeof(uint8_t)*data->interleaveSize*2)) == NULL)
        {
            fprintf(stderr, "Audio data buffer allocation error.\n");
            return PROGRAM_FAILED;
        }
        
        if ( (data->chR = malloc(sizeof(uint8_t)*data->interleaveSize*2)) == NULL)
        {
            fprintf(stderr, "Audio data buffer allocation error.\n");
            return PROGRAM_FAILED;
        }				
        
        printf("Interleave size = %d\n"
               "Sampling rate   = %d\n"
               "Number of CHs   = %d\n"
               "Audio data size = %d\n", data->interleaveSize, data->samplingRate, data->numCh, data->totalAudioDataSize);
        fprintf(stdout, "Converting...");
        fflush(stdout);
        
        writeWAVEHeader(data);
     
        data->channel = data->chL;
    }
    
    data->totalAudioDataCount += audioSize;
    
    /*
     * TODO: 1byte単位ではなく、ある程度まとまった大きさで読み込むように改良する
     */
    for (i=0; i<audioSize; i++)
    {
        // インタリーブサイズ分集まった時の処理
        if (data->interleaveCount == data->interleaveSize)
        {
            data->interleaveCount = 0;
            
            if (data->channel == data->chL)
            {
                data->channel = data->chR;
            }
            else if (data->channel == data->chR)
            {
                data->channel = data->chL;
                writeWAVEData(data);
            }
        }

        fread(&data->channel[data->interleaveCount], 1, 1, data->fpInput);
        data->interleaveCount++;
    }
    DEBUG_PRINT(" ** 1 block finished.\n");
    
    return 0;
}

/*! ///////////////////////////////////////////////////////////////////////////
@brief WAVEファイルのヘッダを書き出す。
@param[in] data PSS2WAVE構造体へのポインタ。
@return なし。
*/ ////////////////////////////////////////////////////////////////////////////
void writeWAVEHeader(PSS2WAVE *data)
{
	int8_t *waveHeader;
	uint32_t fileSize, chunkSize, samplingRate, bytePerSec, totalWaveDataSize;
	uint16_t formatID, numCh, blockSize, bitPerSample;
	
	if ( (waveHeader = malloc(WAVE_HEADER_SIZE)) == NULL)
	{
		fprintf(stderr, "Wave header buffer allocation error.\n");
		exit(EXIT_FAILURE);
	}
	
	fileSize	= INT32_TO_LE(data->totalAudioDataSize +WAVE_HEADER_SIZE -8);
	chunkSize	= INT32_TO_LE((uint32_t)16);
	formatID	= INT16_TO_LE((uint16_t) 1);
	numCh		= INT16_TO_LE(data->numCh);
	samplingRate= INT32_TO_LE(data->samplingRate);
	bytePerSec	= INT32_TO_LE(data->samplingRate * 2 * data->numCh);
	blockSize	= INT16_TO_LE(data->numCh * 2);
	bitPerSample= INT16_TO_LE((uint16_t)16);
	totalWaveDataSize = INT32_TO_LE(data->totalAudioDataSize);

	memcpy(waveHeader, "RIFF",		4); waveHeader+=4;
	memcpy(waveHeader, &fileSize,	4); waveHeader+=4;
	memcpy(waveHeader, "WAVEfmt ",	8); waveHeader+=8;
	memcpy(waveHeader, &chunkSize,	4); waveHeader+=4;
	memcpy(waveHeader, &formatID,	2); waveHeader+=2;
	memcpy(waveHeader, &numCh,		2); waveHeader+=2;
	memcpy(waveHeader, &samplingRate,4);waveHeader+=4;
	memcpy(waveHeader, &bytePerSec,	4);	waveHeader+=4;
	memcpy(waveHeader, &blockSize,	2); waveHeader+=2;
	memcpy(waveHeader, &bitPerSample,2);waveHeader+=2;
	memcpy(waveHeader, "data",		4); waveHeader+=4;
	memcpy(waveHeader, &totalWaveDataSize, 4); waveHeader+=4;
	
	waveHeader-=WAVE_HEADER_SIZE;
    
	fwrite(waveHeader, WAVE_HEADER_SIZE, 1, data->fpOutput);
	fflush(data->fpOutput);
	free(waveHeader);
    
    DEBUG_PRINT(" ** writeWAVEHeader() finished.\n");
}

/* ////////////////////////////////////////////////////////////////////////////
@brief WAVEデータを書き出す。
@param[in] data PSS2WAVE構造体へのポインタ。
@return なし。
*/ ////////////////////////////////////////////////////////////////////////////
void writeWAVEData(PSS2WAVE *data)
{
    int i;
    int16_t wave;
    
    for (i=0; i<data->interleaveSize; i+=2)
    {
        wave = *(short*)&(data->chL[i]);
        fwrite(&wave, 2, 1, data->fpOutput );

        wave = *(short*)&(data->chR[i]);
        fwrite(&wave, 2, 1, data->fpOutput );
    }
    
    fflush(data->fpOutput);

    DEBUG_PRINT("writeWAVEData() finished.\n");
}